@@ -1185,7 +1185,8 @@ class RecordingHandle:
         return True
 
 def tracepoint_supported(tracepoint):
-    output = subprocess.check_output([PERF, 'list']).decode().strip()
+    output = subprocess.check_output(
+            [PERF, 'list', 'tracepoint']).decode().strip()
     for line in output.split('\n'):
         fields = line.split()
         if len(fields) < 3:
@@ -1209,6 +1210,12 @@ def start_recording_perf(handle):
         cmd = [PERF, 'record', '-o', '%s.profile' % handle.file_path]
         handle.perf_profile_pipe = subprocess.Popen(cmd)
 
+def record_source_is_running(record_handle):
+    if record_handle.snapshot_request is not None and damon_stat_avail():
+        return True
+    return poll_target_pids(record_handle.kdamonds) or \
+            _damon.any_kdamond_running()
+
 def start_recording(handle):
     start_recording_perf(handle)
 
@@ -1219,8 +1226,7 @@ def start_recording(handle):
         sleep_time = handle.snapshot_interval_sec
     else:
         sleep_time = 1
-    while (poll_target_pids(handle.kdamonds) or
-           _damon.any_kdamond_running()):
+    while record_source_is_running(handle):
         if handle.add_child_tasks is True:
             _damon.add_commit_vaddr_child_targets(handle.kdamonds)
 
@@ -1309,8 +1315,13 @@ def save_recording_outputs(handle, file_path):
         except:
             # perf might already finished
             pass
-        os.rename('%s.profile' % handle.file_path, '%s.profile' % file_path)
-        os.chmod('%s.profile' % file_path, handle.file_permission)
+        if not os.path.isfile('%s.profile' % handle.file_path):
+            # perf output might not generated due to absence of events.
+            pass
+        else:
+            os.rename('%s.profile' % handle.file_path, '%s.profile' %
+                      file_path)
+            os.chmod('%s.profile' % file_path, handle.file_permission)
 
     if handle.mem_footprint_snapshots is not None:
         save_mem_footprint(
@@ -1659,6 +1670,17 @@ class SnapshotRequest:
         self.total_sz_only = total_sz_only
         self.dont_merge_regions = dont_merge_regions
 
+def damon_stat_avail():
+    param_dir = '/sys/module/damon_stat/parameters/'
+    aggr_interval_us_file = os.path.join(param_dir, 'aggr_interval_us')
+    if not os.path.isfile(aggr_interval_us_file):
+        return False
+    enabled_file = os.path.join(param_dir, 'enabled')
+    with open(enabled_file, 'r') as f:
+        if f.read().strip().lower() == 'n':
+            return False
+    return True
+
 def should_get_snapshot_from_damon_stat(request):
     if request.tried_regions_of is not None:
         return False
@@ -1668,16 +1690,7 @@ def should_get_snapshot_from_damon_stat(request):
     if _damon.any_kdamond_running():
         return False
 
-    param_dir = '/sys/module/damon_stat/parameters/'
-    aggr_interval_us_file = os.path.join(param_dir, 'aggr_interval_us')
-    if not os.path.isfile(aggr_interval_us_file):
-        return False
-    enabled_file = os.path.join(param_dir, 'enabled')
-    with open(enabled_file, 'r') as f:
-        if f.read().strip().lower() == 'n':
-            return False
-
-    return True
+    return damon_stat_avail()
 
 def read_damon_stat_param(param_name):
     file_path = os.path.join('/sys/module/damon_stat/parameters', param_name)
