@@ -142,6 +142,14 @@ def mk_handle(args, kdamonds, monitoring_intervals):
 
     return handle
 
+def for_damon_stat_snapshot(args):
+    if args.snapshot is None:
+        return False
+    request, err = snapshot_requests_from_args(args)
+    if err is not None:
+        return False
+    return _damo_records.should_get_snapshot_from_damon_stat(request)
+
 def main(args):
     global data_for_cleanup
 
@@ -155,7 +163,23 @@ def main(args):
     signal.signal(signal.SIGTERM, sighandler)
 
     # Now the real works
-    if not _damon_args.is_ongoing_target(args):
+    if _damon_args.is_ongoing_target(args):
+        if not _damon.any_kdamond_running():
+            print('DAMON is not turned on')
+            exit(1)
+
+        # TODO: Support multiple kdamonds, multiple contexts
+        monitoring_intervals = data_for_cleanup.orig_kdamonds[
+                0].contexts[0].intervals
+        kdamonds = data_for_cleanup.orig_kdamonds
+    elif for_damon_stat_snapshot(args):
+        monitoring_intervals = _damon.DamonIntervals(
+                sample=5000, aggr=100000, ops_update=60000000,
+                intervals_goal= _damon.DamonIntervalsGoal(
+                    access_bp=400, aggrs=3, min_sample_us=5000,
+                    max_sample_us=10000000))
+        kdamonds = []
+    else:
         err, kdamonds = _damon_args.turn_damon_on(args)
         if err:
             print('could not turn DAMON on (%s)' % err)
@@ -166,15 +190,6 @@ def main(args):
         monitoring_intervals = kdamonds[0].contexts[0].intervals
         now_kdamonds = _damon.current_kdamonds()
         kdamonds[0].pid = now_kdamonds[0].pid
-    else:
-        if not _damon.any_kdamond_running():
-            print('DAMON is not turned on')
-            exit(1)
-
-        # TODO: Support multiple kdamonds, multiple contexts
-        monitoring_intervals = data_for_cleanup.orig_kdamonds[
-                0].contexts[0].intervals
-        kdamonds = data_for_cleanup.orig_kdamonds
 
     record_handle = mk_handle(args, kdamonds, monitoring_intervals)
 
