@@ -8,20 +8,34 @@ import _test_damo_common
 
 _test_damo_common.add_damo_dir_to_syspath()
 
+import _damo_sysinfo
 import _damon
 import _damon_args
 import _damon_sysfs
 
+def set_damon_sysfs_features():
+    avail_features = []
+    for name, support in {
+            'init_regions': True, 'schemes': True, 'schemes_stat_qt_exceed':
+            True, 'init_regions_target_idx': True, 'schemes_prioritization':
+            True, 'schemes_tried_regions': False, 'record': False,
+            'schemes_time_quota': True, 'fvaddr': False, 'paddr': True,
+            'schemes_wmarks': True, 'schemes_size_quota': True,
+            'schemes_stat_succ': True, 'vaddr': True}.items():
+        if not support:
+            continue
+        for feature in _damo_sysinfo.damon_features:
+            if feature.name == name:
+                avail_features.append(feature)
+    _damo_sysinfo.system_info = _damo_sysinfo.SystemInfo(
+            damo_version=None, kernel_version=None,
+            avail_damon_sysfs_features=avail_features,
+            avail_damon_debugfs_features=[])
+
 class TestDamonArgs(unittest.TestCase):
     def test_damon_ctxs_for(self):
         _damon._damon_fs = _damon_sysfs
-        _damon.set_feature_supports({'init_regions': True, 'schemes': True,
-                'schemes_stat_qt_exceed': True, 'init_regions_target_idx':
-                True, 'schemes_prioritization': True, 'schemes_tried_regions':
-                False, 'record': False, 'schemes_quotas': True, 'fvaddr':
-                False, 'paddr': True, 'schemes_wmarks': True,
-                'schemes_speed_limit': True, 'schemes_stat_succ': True,
-                'vaddr': True})
+        set_damon_sysfs_features()
 
         parser = argparse.ArgumentParser()
         _damon_args.set_argparser(
